#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <string.h>
#include <netinet/in.h>
#include <stdlib.h>

#include "config.h"

#include "world.h"
#include "settings.h"

/* These values taken from RFC 854 and RFC 857. */
#define TN_WILL 251
#define TN_WONT 252
#define TN_DO   253
#define TN_DONT 254
#define TN_IAC  255   /* Interpret As Command */
#define TN_SB   250   /* start of subnegotiation */
#define TN_SE   240   /* end of subnegotiaton */

void do_disconnect(world *w);
void abnormal_disconnect(world *w, int err);

server_t *init_server() {
    server_t *server = g_malloc(sizeof(server_t));
    server->status = NO_CONNECTION;
    server->socket = -1;
    server->iotag = -1;
    server->tn_last = 0;
    server->tn_subneg = 0;
    return server;
}

void read_data(world *w, int socket) {
    int err, len;
    unsigned char lbuf[1024];
    unsigned char rbuf[1024];
    unsigned char *ptr,*rptr,*endp;
    server_t *s = w->server;
 
    len = recv(socket, lbuf, sizeof(lbuf) - 2, 0);
    if(len < 1) {
        if(len < 0) {
            if(errno == EAGAIN || errno == EWOULDBLOCK)
                return;                                  
            err = errno;
        } else {
            err = 0;
        }
        abnormal_disconnect(w, err);
        if(settings->reconnect)
            gm_net_connect(w, w->p->hostname, w->p->port);
        return;
    } else {
      rptr = rbuf;
      endp = lbuf + len;
      for (ptr = lbuf; ptr <  endp ; ptr++){
        unsigned char c = *ptr;
        /* handle telnet codes the right way 
         * IAC IAC -> data byte 255
         * IAC WILL/WONT/DO/DONT xxx -> ignore
         * IAC SB xxx ... IAC SE -> ignore
         * IAC other -> ignore
         */
        if (s->tn_last) {
          switch (s->tn_last) {
            case TN_WILL: case TN_WONT: case TN_DO: case TN_DONT:
              s->tn_last = 0;
              break;
            case TN_IAC:
              switch (c) {
                case TN_WILL: case TN_WONT: case TN_DO: case TN_DONT:
                  s->tn_last = c;
                  break;
                case TN_SB:
                  s->tn_subneg = 1; s->tn_last = 0; break;
                case TN_SE:
                  s->tn_subneg = 0; s->tn_last = 0; break;
                case TN_IAC:
                  if (!s->tn_subneg) *rptr++ = c;
                     s->tn_last = 0; break;
                default:
                   s->tn_last = 0;
               }
          }
        }
        else if (c == TN_IAC) s->tn_last = TN_IAC;
        else if (s->tn_subneg) continue;
        else *rptr++ = c;
      }
      gm_world_handle_output_data(w, rbuf, rptr - rbuf);
    }   
}
          
void gm_net_try_to_connect(world *w );

void gm_net_try_connect_next(world *w) {
  if (w->server->current == NULL) 
    w->server->current = w->server->ans;
  else 
    w->server->current = w->server->current->ai_next;

  gm_net_try_to_connect(w);
}

void gm_net_connection_fatal(world *w) {
  gchar *s;
  s =g_strdup_printf("\x1b[31m%% Giving up......\x1b[0m");
  gm_world_println(w, s);
  g_free(s);
  gm_world_disconnected(w);
  do_disconnect(w);
  if(settings->reconnect)
	 gm_net_connect(w, w->p->hostname, w->p->port);
}

void gm_net_connection_failed(world *w,int error) {
  gchar *s;

  gdk_input_remove(w->server->iotag); w->server->iotag = -1;
  if (w->server->socket != -1) close(w->server->socket);
  s =g_strdup_printf("\x1b[31m%% Connecting failed: %s (error %d)\x1b[0m",
                            strerror(error), error);
  gm_world_println(w, s);
  g_free(s);
  gm_net_try_connect_next(w);
}

void gm_net_connection_succeeded(world *w) {
  if (fcntl(w->server->socket,F_SETFL,O_NONBLOCK) < 0) 
    return gm_net_connection_failed(w,errno);
  gdk_input_remove(w->server->iotag);
  w->server->iotag =  gdk_input_add(w->server->socket,
                      GDK_INPUT_READ|GDK_INPUT_EXCEPTION,
                      (GdkInputFunction) read_data, w);
  w->server->status = CONNECTED;
  freeaddrinfo(w->server->ans);
  w->server->ans = NULL;
  gm_world_connected(w);
}

void gm_connecting_callback(world *w, int socket,GdkInputCondition condition) {
  int error;
  int len = sizeof(error);

  if (getsockopt(w->server->socket,SOL_SOCKET,SO_ERROR,&error,&len) < 0)
        return gm_net_connection_failed(w,errno);
  if (error != 0) return gm_net_connection_failed(w,error); 
  gm_net_connection_succeeded(w);
}

void gm_net_try_to_connect(world *w ) {
  struct addrinfo *tmpaddr = w->server->current;
  char *s,name[NI_MAXHOST],portname[NI_MAXSERV];

  w->server->socket = -1;

  if (w->server->current == NULL) return gm_net_connection_fatal(w);

  getnameinfo(tmpaddr->ai_addr, tmpaddr->ai_addrlen,
              name,sizeof(name),portname,sizeof(portname),
                      NI_NUMERICHOST | NI_NUMERICSERV);
  s = g_strdup_printf("\x1b[32m%% Trying %s port %s\x1b[0m",name,portname);
  gm_world_println(w, s);
  g_free(s);

  w->server->socket =  
    socket(tmpaddr->ai_family,tmpaddr->ai_socktype,tmpaddr->ai_protocol);

  if (w->server->socket < 0) return gm_net_connection_failed(w,errno);

  if (fcntl(w->server->socket,F_SETFL,O_NONBLOCK) < 0) 
    return gm_net_connection_failed(w,errno);

  if (connect(w->server->socket, 
      w->server->current->ai_addr, 
      w->server->current->ai_addrlen) < 0
      && errno != EINPROGRESS) {
    return gm_net_connection_failed(w,errno);
  }
  w->server->iotag = 
    gdk_input_add(w->server->socket, 
                  GDK_INPUT_WRITE|GDK_INPUT_EXCEPTION,
                  (GdkInputFunction) gm_connecting_callback, w);
} 

void gm_net_connect(world *w, const char *hostname, int port) {
    char *s,sport[NI_MAXSERV];
    struct addrinfo req;
    
    if(debug) printf("NET: Connecting...\n");
    if(!hostname || !hostname[0] || port < 1)
        return;

    if(w->server) {
        if(debug) printf("NET: Disconnecting current\n");
        do_disconnect(w);
    } else {
        if(debug) printf("NET: Creating new server\n");
        w->server = init_server();
    }
    
    s = g_strdup_printf("\x1b[32m%% Connecting to %s:%d...\x1b[0m",
                        hostname, port);
    gm_world_println(w, s);
    g_free(s);

    memset(&req,0,sizeof(req));
    req.ai_flags = 0;
    req.ai_family = AF_UNSPEC;
    req.ai_socktype = SOCK_STREAM;
    req.ai_protocol =IPPROTO_TCP;
    snprintf(sport,sizeof(sport),"%d",port);
    if (getaddrinfo(hostname,sport,&req,&(w->server->ans)) != 0) {
      s = g_strdup_printf("\x1b[31m%% Hostname lookup failed \x1b[0m");
      gm_world_println(w, s);
      g_free(s);
      return;
    }

    w->server->current = NULL;
    w->server->status = CONNECTING;
    w->server->tn_last = 0;
    w->server->tn_subneg = 0;
    gm_world_connecting(w);
    gm_net_try_connect_next(w);
}

void 
do_disconnect(world *w) {
  if(!w->server) return;
  if (w->server->status != NO_CONNECTION) {
    if (w->server->iotag != -1) {
      gdk_input_remove(w->server->iotag);
      w->server->iotag = -1;
    }
    close(w->server->socket);
    if (w->server->status == CONNECTING) {
        freeaddrinfo(w->server->ans);
        w->server->ans = NULL;
    }
  }
   w->server->status = NO_CONNECTION;
}

void gm_net_disconnect(world *w) {
    do_disconnect(w);
    gm_world_println(w, "\x1b[32m% Disconnected\x1b[0m");
    gm_world_disconnected(w);
}

void abnormal_disconnect(world *w, int err) {
    char *s;
    do_disconnect(w);
    if(err) {
        s = g_strdup_printf("\x1b[31m%% Connection lost: %s (error %d)\x1b[0m",
                         strerror(err), err);
        gm_world_println(w, s);
        g_free(s);
    } else {
        gm_world_println(w, "\x1b[31m% Connection lost\x1b[0m");
    }
    gm_world_disconnected(w);
}


void gm_net_send(world *w, const unsigned char *text, int length) {
    const unsigned char *ptr,*pptr,*eptr;
    if(w->server) {
        if(w->server->status == CONNECTED) {
          ptr = pptr = text;
          eptr = text + length;
          for (; ptr < eptr ; ptr++) { 
            if (*ptr == TN_IAC) {
              send(w->server->socket, pptr, ptr - pptr + 1, 0);
              send(w->server->socket,ptr,1,0);
              pptr = ptr + 1;
            }
          }
          send(w->server->socket, pptr, ptr - pptr, 0);
          return;
        }
        if(w->server->status == CONNECTING) {
            gm_world_println(w, "% Still connecting ...");
            return;
        }
    }
    gm_world_println(w, "\x1b[31m% Not connected !\x1b[0m");
}

